import random
import warnings
from functools import lru_cache

import numpy as np
import torch


#@lru_cache(maxsize=32)
def get_vanilla_image_gradient(model, inpt, err_fn, abs=False):
    if isinstance(model, torch.nn.Module):
        model.zero_grad()
    inpt = inpt.detach()
    inpt.requires_grad = True

    output = model(inpt)

    err = err_fn(output)
    err.backward()

    grad = inpt.grad.detach()

    if isinstance(model, torch.nn.Module):
        model.zero_grad()

    if abs:
        grad = torch.abs(grad)
    return grad.detach()


#@lru_cache(maxsize=32)
def get_guided_image_gradient(model: torch.nn.Module, inpt, err_fn, abs=False):
    def guided_relu_hook_function(module, grad_in, grad_out):
        if isinstance(module, (torch.nn.ReLU, torch.nn.LeakyReLU)):
            return (torch.clamp(grad_in[0], min=0.0),)

    model.zero_grad()

    ### Apply hooks
    hook_ids = []
    for mod in model.modules():
        hook_id = mod.register_backward_hook(guided_relu_hook_function)
        hook_ids.append(hook_id)

    inpt = inpt.detach()
    inpt.requires_grad = True

    output = model(inpt)

    err = err_fn(output)
    err.backward()

    grad = inpt.grad.detach()

    model.zero_grad()
    for hooks in hook_ids:
        hooks.remove()

    if abs:
        grad = torch.abs(grad)
    return grad.detach()


#@lru_cache(maxsize=32)
def get_smooth_image_gradient(model, inpt, err_fn, n_runs=20, eps=0.1, grad_type="vanilla"):
    grads = []
    for i in range(n_runs):
        inpt = inpt + torch.randn(inpt.size()).to(inpt.device) * eps
        if grad_type == "vanilla":
            single_grad = get_vanilla_image_gradient(model, inpt, err_fn)
        elif grad_type == "guided":
            single_grad = get_guided_image_gradient(model, inpt, err_fn)
        else:
            warnings.warn("This grad_type is not implemented yet")
            single_grad = torch.zeros_like(inpt)
        grads.append(torch.abs(single_grad))

    grad = torch.mean(torch.stack(grads), dim=0)
    return grad.detach()


def update_model(original_model, update_dict, exclude_layers=(), do_warnings=True):
    # also allow loading of partially pretrained net
    model_dict = original_model.state_dict()

    # 1. Give warnings for unused update values
    unused = set(update_dict.keys()) - set(exclude_layers) - set(model_dict.keys())
    not_updated = set(model_dict.keys()) - set(exclude_layers) - set(update_dict.keys())
    if do_warnings:
        for item in unused:
            warnings.warn("Update layer {} not used.".format(item))
        for item in not_updated:
            warnings.warn("{} layer not updated.".format(item))

    # 2. filter out unnecessary keys
    update_dict = {k: v for k, v in update_dict.items() if
                   k in model_dict and k not in exclude_layers}

    # 3. overwrite entries in the existing state dict
    model_dict.update(update_dict)

    # 4. load the new state dict
    original_model.load_state_dict(model_dict)


def set_seed(seed):
    """Sets the seed"""
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    # if torch.cuda.is_available():
    #     torch.cuda.manual_seed_all(seed)



import atexit
import os
from collections import defaultdict

from tensorboardX import SummaryWriter

from trixi.logger.plt.numpyseabornplotlogger import NumpySeabornPlotLogger
from trixi.logger.abstractlogger import convert_params
from trixi.util.util import np_make_grid


class TensorboardXLogger(NumpySeabornPlotLogger):
    """Logger that uses tensorboardX to log to Tensorboard."""

    def __init__(self, target_dir, *args, **kwargs):

        super(TensorboardXLogger, self).__init__(*args, **kwargs)

        os.makedirs(target_dir, exist_ok=True)

        self.writer = SummaryWriter(target_dir)
        self.val_dict = defaultdict(int)

        atexit.register(self.writer.close)

    def show_image(self, image, name="Image", counter=None, **kwargs):
        """
        Sends an image to tensorboard.

        Args:
            image (np.narray/torch.tensor): Image array/tensor which will be sent
            name (str): Identifier for the image
            counter (int): Global step value
        """

        if counter is not None:
            self.val_dict["{}-image".format(name)] = counter
        else:
            self.val_dict["{}-image".format(name)] += 1

        self.writer.add_image(name, image, global_step=self.val_dict["{}-image".format(name)])

    def show_images(self, images, name="Images", counter=None, **kwargs):
        """
        Sends multiple images to tensorboard.

        Args:
            image (np.narray/torch.tensor): Image array/tensor which will be sent (NxCxHxW)
            name (str): Identifier for the images
            counter (int): Global step value
        """

        if counter is not None:
            self.val_dict["{}-image".format(name)] = counter
        else:
            self.val_dict["{}-image".format(name)] += 1

        self.writer.add_images(name, images, global_step=self.val_dict["{}-image".format(name)])

    @convert_params
    def show_value(self, value, name="Value", counter=None, tag=None, **kwargs):
        """
        Sends a scalar value to tensorboard.

        Args:
            value (numeric): Value to be sent
            name (str): Identifier for the value
            counter (int): Global step value
            tag (str): Identifier for the frame (values with the same tag will be shown in the same graph)
        """

        if tag is None:
            key = name + "-" + name
        else:
            key = tag + "-" + name

        if counter is not None:
            self.val_dict["{}-image".format(key)] = counter
        else:
            self.val_dict["{}-image".format(key)] += 1

        if tag is not None:
            self.writer.add_scalars(tag, {name: value}, global_step=self.val_dict["{}-image".format(key)])
            self.writer.scalar_dict = {}
        else:
            self.writer.add_scalar(name, value, global_step=self.val_dict["{}-image".format(key)])

    def show_text(self, text, name="Text", counter=None, **kwargs):
        """
        Sends text to tensorboard.

        Args:
            text (str): Text to be sent
            name (str): Identifier for the text
            counter (int): Global step value
        """

        if counter is not None:
            self.val_dict["{}-text".format(name)] = counter
        else:
            self.val_dict["{}-text".format(name)] += 1

        self.writer.add_text(name, text, global_step=self.val_dict["{}-text".format(name)])

    @convert_params
    def show_image_grid(self, image_array, name="Image-Grid", counter=None, nrow=8, padding=2,
                        normalize=False, range=None, scale_each=False, pad_value=0,
                        *args, **kwargs):
        """
        Sends an array of images to tensorboard as a grid. Like :meth:`.show_image`, but generates
        image grid before.

        Args:
            image_array (np.narray/torch.tensor): Image array/tensor which will be sent as an image grid
            name (str): Identifier for the image grid
            counter (int): Global step value
            nrow (int): Items per row in grid
            padding (int): Padding between images in grid
            normalize (bool): Normalize images in grid
            range (tuple): Tuple (min, max), so images will be normalized to this range
            scale_each (bool): If True, each image will be normalized separately instead of using
                min and max of whole tensor
            pad_value (float): Fill padding with this value
        """

        image_args = dict(nrow=nrow,
                          padding=padding,
                          normalize=normalize,
                          range=range,
                          scale_each=scale_each,
                          pad_value=pad_value)

        if counter is not None:
            self.val_dict["{}-image".format(name)] = counter
        else:
            self.val_dict["{}-image".format(name)] += 1

        grid = np_make_grid(image_array, **image_args)
        self.writer.add_image(tag=name, img_tensor=grid, global_step=self.val_dict["{}-image".format(name)])
        self.val_dict[name] += 1

    @convert_params
    def show_barplot(self, array, name="barplot", counter=None, *args, **kwargs):
        """
        Sends a barplot to tensorboard.

        Args:
            array (np.array/torch.tensor): array of shape NxM where N is the number of rows and M is the number of elements in the row.
            name (str): The name of the figure
            counter (int): Global step value to record

        """

        if counter is not None:
            self.val_dict["{}-figure".format(name)] = counter
        else:
            self.val_dict["{}-figure".format(name)] += 1

        figure = super().show_barplot(array, name, *args, **kwargs)
        self.writer.add_figure(tag=name, figure=figure, global_step=self.val_dict["{}-figure".format(name)])

    @convert_params
    def show_lineplot(self, y_vals, x_vals=None, name="lineplot", counter=None, *args, **kwargs):
        """
        Sends a lineplot to tensorboard.

        Args:
            y_vals (np.array/torch.tensor): Array of shape MxN , where M is the number of points and N is the number of different line
            x_vals (np.array/torch.tensor): Has to have the same shape as Y: MxN. For each point in Y it gives the corresponding X value (if
                not set the points are assumed to be equally distributed in the interval [0, 1])
            name (str): The name of the figure
            counter (int): Global step value to record

        """

        if counter is not None:
            self.val_dict["{}-figure".format(name)] = counter
        else:
            self.val_dict["{}-figure".format(name)] += 1

        figure = super().show_lineplot(y_vals, x_vals, name, *args, **kwargs)
        self.writer.add_figure(tag=name, figure=figure, global_step=self.val_dict["{}-figure".format(name)])

    @convert_params
    def show_scatterplot(self, array, name="scatterplot", counter=None, *args, **kwargs):
        """
        Sends a scatterplot to tensorboard.

        Args:
            array (np.array/torch.tensor): An array with size N x dim, where each element i \in N` at X[i] results in a 2D
                (if dim = 2) or 3D (if dim = 3) point.
            name (str): The name of the figure
            counter (int): Global step value to record

        """

        if counter is not None:
            self.val_dict["{}-figure".format(name)] = counter
        else:
            self.val_dict["{}-figure".format(name)] += 1

        figure = super().show_scatterplot(array, name, *args, **kwargs)
        self.writer.add_figure(tag=name, figure=figure, global_step=self.val_dict["{}-figure".format(name)])

    @convert_params
    def show_piechart(self, array, name="piechart", counter=None, *args, **kwargs):
        """
        Sends a piechart tensorboard.

        Args:
            array (np.array/torch.tensor): Array of positive integers. Each integer will be
                presented as a part of the pie (with the total as the sum of all integers)
            name (str): The name of the figure
            counter (int): Global step value to record

        """

        if counter is not None:
            self.val_dict["{}-figure".format(name)] = counter
        else:
            self.val_dict["{}-figure".format(name)] += 1

        figure = super().show_piechart(array, name, *args, **kwargs)
        self.writer.add_figure(tag=name, figure=figure, global_step=self.val_dict["{}-figure".format(name)])

    def show_embedding(self, tensor, labels=None, name='default', label_img=None, counter=None,
                       *args, **kwargs):
        """
        Displays an embedding of a tensor (for more details see tensorboardX)

        Args:
            tensor (torch.tensor/np.array): Tensor to be embedded and then displayed
            labels (list): List of labels, each element will be converted to string
            name (str): The name for the embedding
            label_img (torch.tensor): Images to be displayed at the embedding points
            counter (int):  Global step value to record

        """

        if counter is not None:
            self.val_dict["{}-embedding".format(name)] = counter
        else:
            self.val_dict["{}-embedding".format(name)] += 1

        self.writer.add_embedding(mat=tensor, metadata=labels, label_img=label_img, tag=name, global_step=self.val_dict["{}-embedding".format(name)])

    def show_histogram(self, array, name="Histogram", counter=None, *args, **kwargs):
        """
        Plots a histogram in the tensorboard histrogram plugin

        Args:
            array (torch.tensor/np.array): Values to build histogram
            name (str): Data identifier
            counter (int):  Global step value to record

        """

        if counter is not None:
            self.val_dict["{}-histogram".format(name)] = counter
        else:
            self.val_dict["{}-histogram".format(name)] += 1

        self.writer.add_histogram(tag=name, values=array, global_step=self.val_dict["{}-histogram".format(name)])

    def show_pr_curve(self, tensor, labels, name="pr-curve", counter=None, *args, **kwargs):
        """
        Displays a precision recall curve given a tensor with scores and the corresponding labels

        Args:
            tensor (torch.tensor/np.array): Tensor with scores (e.g class probabilities)
            labels (list): Labels of the samples to which the scores match
            name (str): The name of the plot
            counter (int): Global step value
        """

        if counter is not None:
            self.val_dict["{}-pr-curve".format(name)] = counter
        else:
            self.val_dict["{}-pr-curve".format(name)] += 1

        self.writer.add_pr_curve(tag=name, labels=labels, predictions=tensor, global_step=self.val_dict["{}-pr-curve".format(name)])

    def close(self):
        self.writer.close()



import os
import tempfile
import unittest
import shutil
import time

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from trixi.logger.experiment import PytorchExperimentLogger


class TestPytorchExperimentLogger(unittest.TestCase):

    def setUp(self):
        self.test_dir = tempfile.gettempdir()
        self.experimentLogger = PytorchExperimentLogger(exp_name="test",
                                                        base_dir=self.test_dir,
                                                        folder_format="{experiment_name}")

    def tearDown(self):
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir, ignore_errors=True)

    def test_show_image(self):
        image = np.random.random_sample((3, 128, 128))
        tensor = torch.from_numpy(image)
        self.experimentLogger.show_image(tensor, "image")

        time.sleep(1)
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.img_dir, "image.png")),
                        "Show image could not create image")

    def test_show_images(self):
        images = np.random.random_sample((4, 3, 128, 128))
        tensors = torch.from_numpy(images)
        self.experimentLogger.show_images(tensors, "image")

        time.sleep(1)
        self.assertTrue(len(os.listdir(self.experimentLogger.img_dir)) > 3,
                        "Show images could not create multiple images")

    def test_show_image_grid(self):
        images = np.random.random_sample((4, 3, 128, 128))
        tensor = torch.from_numpy(images)
        self.experimentLogger.show_image_grid(tensor, "image_grid")

        time.sleep(1)
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.img_dir, "image_grid.png")),
                        "Show image grid could not create image grid from tensor")

    def test_show_image_grid_heatmap(self):
        images = np.random.random_sample((4, 3, 128, 128))
        tensor = torch.from_numpy(images)
        self.experimentLogger.show_image_grid_heatmap(tensor, name="image_grid_heatmap")

        time.sleep(1)
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.img_dir, "image_grid_heatmap.png")),
                        "Show image grid could not create image grid from tensor")

    def test_show_barplot(self):
        tensor = torch.from_numpy(np.random.random_sample(5))
        self.experimentLogger.show_barplot(tensor, "barplot")
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.plot_dir, "barplot.png")),
                        "Show barplot could not create barplot")

    def test_net_save_and_load_model(self):
        net = Net()
        self.experimentLogger.save_model(name="model", model=net)
        # give some time before loading, so saving has finished
        time.sleep(1)
        net2 = Net()
        diff_conv1 = np.abs(np.sum(net.conv1.bias.detach().numpy() - net2.conv1.bias.detach().numpy()))
        self.assertTrue(diff_conv1 > 1e-3, "conv1 bias values have not been initialized differently")
        self.experimentLogger.load_model(name="model", model=net2)
        time.sleep(2)

        np.testing.assert_allclose(net.conv1.bias.detach().numpy(), net2.conv1.bias.detach().numpy(),
                                   err_msg="loading model did not restore model values")

    def test_net_save_and_load_checkpoint(self):
        net = Net()
        self.experimentLogger.save_checkpoint(name="checkpoint", net=net)
        # give some time before loading, so saving has finished
        time.sleep(1)
        net2 = Net()
        diff_conv1 = np.abs(np.sum(net.conv1.bias.detach().numpy() - net2.conv1.bias.detach().numpy()))
        self.assertTrue(diff_conv1 > 1e-3, "conv1 bias values have not been initialized differently")
        self.experimentLogger.load_last_checkpoint(dir=self.experimentLogger.checkpoint_dir, net=net2)
        np.testing.assert_allclose(net.conv1.bias.detach().numpy(), net2.conv1.bias.detach().numpy(),
                                   err_msg="loading checkpoint did not restore model values")

    def test_net_save_and_load_checkpoint_with_optimizer(self):
        self._test_load_save_checkpoint(test_with_cuda=False)

    def test_net_save_and_load_checkpoint_with_optimizer_with_cuda(self):
        if torch.cuda.is_available():
            self._test_load_save_checkpoint(test_with_cuda=True)

    def _test_load_save_checkpoint(self, test_with_cuda):
        # small testing net
        net = Net()
        # fake values for fake step
        random_input = torch.from_numpy(np.random.randn(28 * 28).reshape((1, 1, 28, 28))).float()
        fake_labels = torch.from_numpy(np.array([2])).long()

        if test_with_cuda:
            net.cuda()
            random_input, fake_labels = random_input.cuda(), fake_labels.cuda()

        criterion = torch.nn.CrossEntropyLoss()
        optimizer = torch.optim.Adam(net.parameters(), lr=1e-1, eps=1e-7, weight_decay=1e-4, betas=(0.3, 0.4))

        # do a fake forward/backward/step pass
        pred = net(random_input)
        err = criterion(pred, fake_labels)
        err.backward()
        optimizer.step()

        # save and load network and optimizer
        self.experimentLogger.save_checkpoint(name="checkpoint", net=net, optimizer=optimizer)
        # give some time before loading, so saving has finished
        time.sleep(1)

        loaded_network = Net()
        if test_with_cuda:
            loaded_network.cuda()

        loaded_optimizer = torch.optim.Adam(loaded_network.parameters(), lr=1e-1)
        self.experimentLogger.load_last_checkpoint(dir=self.experimentLogger.checkpoint_dir, net=loaded_network,
                                                   optimizer=loaded_optimizer)

        saved_pg = optimizer.param_groups[0]
        loaded_pg = loaded_optimizer.param_groups[0]

        # assert optimizer values have been correctly saved/loaded
        self.assertAlmostEqual(saved_pg["lr"], loaded_pg["lr"], "learning rate not correctly saved")
        self.assertAlmostEqual(saved_pg["eps"], loaded_pg["eps"], "eps not correctly saved")
        self.assertAlmostEqual(saved_pg["weight_decay"], loaded_pg["weight_decay"], "weight_decay not correctly saved")
        self.assertAlmostEqual(saved_pg["betas"][0], loaded_pg["betas"][0], "beta 0 not correctly saved")
        self.assertAlmostEqual(saved_pg["betas"][1], loaded_pg["betas"][1], "beta 1 not correctly saved")
        for saved_parameter, loaded_parameter in zip(saved_pg["params"], loaded_pg["params"]):
            np.testing.assert_allclose(saved_parameter.detach().cpu().numpy(), loaded_parameter.detach().cpu().numpy(),
                                       err_msg="loading checkpoint did not restore optimizer param values")

        # do a fake forward/backward/step pass with loaded stuff
        pred = loaded_network(random_input)
        err = criterion(pred, fake_labels)
        err.backward()
        loaded_optimizer.step()  # this could raise an error, which is why there is no assert but just the execution

    def test_print(self):
        text = "\nTest 4 fun: zD ;-D 0o"
        self.experimentLogger.print(text)
        log_text = ""
        with open(os.path.join(self.experimentLogger.log_dir, "default.log"), 'r') as log_file:
            log_text = log_file.read()
        self.assertTrue(text in log_text)

    def test_get_roc_curve(self):
        array = np.random.random_sample(100)
        labels = np.random.choice((0, 1), 100)

        tpr, fpr = self.experimentLogger.get_roc_curve(array, labels)
        self.assertTrue(np.all(tpr >= 0) and np.all(tpr <= 1) and np.all(fpr >= 0) and np.all(fpr <= 1),
                        "Got an invalid tpr, fpr")

    def test_get_pr_curve(self):
        array = np.random.random_sample(100)
        labels = np.random.choice((0, 1), 100)

        precision, recall = self.experimentLogger.get_pr_curve(array, labels)
        self.assertTrue(np.all(precision >= 0) and np.all(precision <= 1)
                        and np.all(recall >= 0) and np.all(recall <= 1),
                        "Got an invalid precision, recall")

    def test_get_classification_metric(self):
        array = np.random.random_sample(100)
        labels = np.random.choice((0, 1), 100)

        vals, tags = self.experimentLogger.get_classification_metrics(array, labels,
                                                                      metric=("roc-auc", "pr-score"))

        self.assertTrue("roc-auc" in tags and "pr-score" in tags, "Did not get all classification metrics")
        self.assertTrue(vals[0] >= 0 and vals[0] <= 1
                        and vals[1] >= 0 and vals[1] <= 1,
                        "Got an invalid classification metric values")

    def test_show_image_gradient(self):

        net = Net()
        random_input = torch.from_numpy(np.random.randn(28 * 28).reshape((1, 1, 28, 28))).float()
        fake_labels = torch.from_numpy(np.array([2])).long()
        criterion = torch.nn.CrossEntropyLoss()

        err_fn = lambda x: criterion(x, fake_labels)

        self.experimentLogger.show_image_gradient("grads-vanilla", model=net, inpt=random_input, err_fn=err_fn,
                                                  grad_type="vanilla")
        time.sleep(1)
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.img_dir, "grads-vanilla.png")),
                        "Could not get vanilla gradients")

        self.experimentLogger.show_image_gradient("grads-svanilla", model=net, inpt=random_input, err_fn=err_fn,
                                                  grad_type="smooth-vanilla")
        time.sleep(1)
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.img_dir, "grads-svanilla.png")),
                        "Could not get vanilla gradients")

        self.experimentLogger.show_image_gradient("grads-guided", model=net, inpt=random_input, err_fn=err_fn,
                                                  grad_type="guided")
        time.sleep(1)
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.img_dir, "grads-guided.png")),
                        "Could not get vanilla gradients")

        self.experimentLogger.show_image_gradient("grads-sguided", model=net, inpt=random_input, err_fn=err_fn,
                                                  grad_type="smooth-guided")
        time.sleep(1)
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.img_dir, "grads-sguided.png")),
                        "Could not get vanilla gradients")


class Net(nn.Module):
    """
    Small network to test save/load functionality
    """

    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


if __name__ == '__main__':
    unittest.main()



from trixi.logger.abstractlogger import AbstractLogger
from trixi.logger.combinedlogger import CombinedLogger
from trixi.logger.plt.numpyseabornplotlogger import NumpySeabornPlotLogger
from trixi.logger.file.numpyplotfilelogger import NumpyPlotFileLogger
from trixi.logger.file.textfilelogger import TextFileLogger
from trixi.logger.visdom.numpyvisdomlogger import NumpyVisdomLogger
from trixi.logger.experiment.experimentlogger import ExperimentLogger

try:
    from trixi.logger.experiment.pytorchexperimentlogger import PytorchExperimentLogger
    from trixi.logger.file.pytorchplotfilelogger import PytorchPlotFileLogger
    from trixi.logger.visdom.pytorchvisdomlogger import PytorchVisdomLogger
except ImportError as e:
    import warnings
    warnings.warn(ImportWarning("Could not import Pytorch related modules:\n%s"
        % e.msg))

try:
    from trixi.logger.message.telegrammessagelogger import TelegramMessageLogger
except ImportError as e:
    import warnings
    warnings.warn(ImportWarning("Could not import Telegram related modules:\n%s"
        % e.msg))



try:
    from trixi.logger.message.telegrammessagelogger import TelegramMessageLogger
except ImportError as e:
    import warnings
    warnings.warn(ImportWarning("Could not import telegram related modules:\n%s"
        % e.msg))
